# -- STL Imports --
from abc import ABC as _Namespace
from typing import TypeAlias

# -- Package Imports --
from parasite.any import Any_
from parasite.array import Array
from parasite.boolean import Boolean
from parasite.null import Null
from parasite.number import Number
from parasite.object import Object
from parasite.string import String
from parasite.variant import Variant
from parasite.never import Never


class p(_Namespace):
    """
    sudo-namespace for all parasite types. Makes it easier to import and call them. Tries to mimic
    the behavior of the ``z`` object imported from ``zod`` library in JavaScript.

    Example usage::

        >>> from parasite import p
        >>>
        >>> schema = p.obj({
        ...     "name": p.string().required(),
        ...     "age": p.number().integer().min(0).optional(),
        ... }).strip()
        >>>
        >>> data = {
        ...     "name": "John Doe",
        ...     "age": 42,
        ...     "extra": "This will be stripped",
        ... }
        >>>
        >>> schema.parse(data)
        {'name': 'John Doe', 'age': 42}
        >>>
        >>> schema.parse({})
        ValidationError: Missing required key: 'name'
    """
    any: TypeAlias = Any_
    null: TypeAlias = Null
    number: TypeAlias = Number
    string: TypeAlias = String
    boolean: TypeAlias = Boolean
    never: TypeAlias = Never
    variant: TypeAlias = Variant
    obj: TypeAlias = Object
    array: TypeAlias = Array



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass, field
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some
from rusttypes.result import Result, Err, Ok

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Variant(ParasiteType[Any]):
    """
    Parasite type for representing variant values.

    Inheritance:
        ParasiteType[Any]

    Args:
        _f_optional (bool): Whether the value is optional. Default: False
        _f_nullable (bool): Whether the value can be None. Default: False
        _m_variants (set[ParasiteType]): The variants of the variant. Default: {}
    """
    # NOTE: do not move this attribute, this has to be first in the class, as it will break, reading
    # variants from list functionality
    _m_variants: list[ParasiteType] = field(default_factory = lambda: [])

    _f_optional: bool = False
    _f_nullable: bool = False

    def optional(self) -> Variant:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Returns:
            Variant: The updated instance of the class.
        """
        self._f_optional = True
        return self

    def required(self) -> Variant:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Returns:
            Variant: The updated instance of the class.
        """
        self._f_optional = False
        return self

    def nullable(self) -> Variant:
        """
        Set the value to be nullable.

        Returns:
            Variant: The updated instance of the class.
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Variant:
        """
        Set the value to be non-nullable.

        Returns:
            Variant: The updated instance of the class.
        """
        self._f_nullable = False
        return self

    def add_variant(self, variant: ParasiteType) -> Variant:
        """
        Add a variant to the variant.

        Args:
            variant (ParasiteType): The variant to add.

        Returns:
            Variant: The updated instance of the class.
        """
        self._m_variants.append(variant)
        return self

    def rm_variant(self, variant: ParasiteType) -> Variant:
        """
        Remove a variant from the variant.

        Throws:
            ValueError: If the variant is not found in the variant.

        Args:
            variant (ParasiteType): The variant to remove.

        Returns:
            Variant: The updated instance of the class.
        """
        try:
            self._m_variants.remove(variant)
        except ValueError as exc:
            raise ValueError(f"Variant {variant!r} not found in {self!r}") from exc

        return self

    def rm_variant_safe(self, variant: ParasiteType) -> Result[Variant, ValueError]:
        """
        Remove a variant from the variant.

        Returns:
            Optional[ParasiteType]: The removed variant.
        """
        try:
            self.rm_variant(variant)
            return Ok(self)

        except ValueError as exc:
            return Err(exc)

    def parse(self, obj: Any) -> Any:
        for variant in self._m_variants:
            try:
                return variant.parse(obj)

            except ValidationError:
                continue

        raise ValidationError(f"object has to be one of {self._m_variants!r}, but is '{obj!r}'")

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[Any | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # if key is found, just package ``parse(..)`` it into a Some
            if value is not None:
                return Some(self.parse(value))

            # if value is None, check if the value is nullable
            if self._f_nullable:
                return Some(None)

            raise ValidationError(f"key '{key}' cannot be None")

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
from typing import Any, TypeAlias, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""

Numerical: TypeAlias = int | float


@dataclass
class Number(ParasiteType[Numerical]):
    """
    Parasite type for representing numerical values. This type can parse both integers and floats.

    Note:
        Please use ``p.number()`` instead of instantiating this class directly. ``p`` can be
        imported with::

            from parasite import p
            schema = p.number()
            ...
    """

    _f_optional: bool = False   # Wether the value is optional
    _f_nullable: bool = False   # Wether the value can be None
    _f_integer: bool = False   # Wether the value has to be an integer

    _f_lt: bool = False   # Wether the value has to be less than a certain value
    _f_lte: bool = False   # Wether the value has to be less than or equal to a certain value
    _f_gt: bool = False   # Wether the value has to be greater than a certain value
    _f_gte: bool = False   # Wether the value has to be greater than or equal to a certain value

    _m_ul: Numerical | None = None   # Upper limit for the value
    _m_ll: Numerical | None = None   # Lower limit for the value

    def __init__(self) -> None:
        pass

    def optional(self) -> Number:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Returns:
            Number: modified instance
        """
        self._f_optional = True
        return self

    def required(self) -> Number:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Returns:
            Number: modified instance
        """
        self._f_optional = False
        return self

    def nullable(self) -> Number:
        """
        Makes the value nullable, when parsing with ``_find_and_parse(..)``. Has no effect on
        ``parse(..)``. Inverse of ``non_nullable(..)``.

        Returns:
            Number: modified instance
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Number:
        """
        Makes the value non-nullable, when parsing with ``_find_and_parse(..)``. Has no effect on
        ``parse(..)``. Default behavior. Inverse of ``nullable(..)``.

        Returns:
            Number: modified instance
        """
        self._f_nullable = False
        return self

    def integer(self) -> Number:
        """
        Makes the value an integer, when parsing with ``_find_and_parse(..)``. Has no effect on
        ``parse(..)``. Inverse of ``float(..)``.

        Returns:
            Number: modified instance
        """
        self._f_integer = True
        return self

    def float(self) -> Number:
        """
        Makes the value a float, when parsing with ``_find_and_parse(..)``. Has no effect on
        ``parse(..)``. Default behavior. Inverse of ``integer(..)``.

        Returns:
            Number: modified instance
        """
        self._f_integer = False
        return self

    def gt(self, value: Numerical) -> Number:
        """
        Sets the lower limit for the value. The value has to be greater than the specified value.

        Args:
            value (Numerical): lower limit for the value

        Returns:
            Number: modified instance
        """
        self._f_gte = False
        self._f_gt = True
        self._m_ll = value
        return self

    def gte(self, value: Numerical) -> Number:
        """
        Sets the lower limit for the value. The value has to be greater than or equal to the
        specified value.

        Args:
            value (Numerical): lower limit for the value

        Returns:
            Number: modified instance
        """
        self._f_gt = False
        self._f_gte = True
        self._m_ll = value
        return self

    def positive(self) -> Number:
        """
        Sets the lower limit for the value. The value has to be greater than 0.

        Returns:
            Number: modified instance
        """
        return self.gt(0)

    def not_negative(self) -> Number:
        """
        Sets the lower limit for the value. The value has to be greater than or equal to 0.

        Returns:
            Number: modified instance
        """
        return self.gte(0)

    def min(self, value: Numerical) -> Number:
        """
        Sets the lower limit for the value. The value has to be greater than or equal to the
        specified value.

        Args:
            value (Numerical): lower limit for the value

        Returns:
            Number: modified instance
        """
        return self.gte(value)

    def lt(self, value: Numerical) -> Number:
        """
        Sets the upper limit for the value. The value has to be less than the specified value.

        Args:
            value (Numerical): upper limit for the value

        Returns:
            Number: modified instance
        """
        self._f_lte = False
        self._f_lt = True
        self._m_ul = value
        return self

    def lte(self, value: Numerical) -> Number:
        """
        Sets the upper limit for the value. The value has to be less than or equal to the specified
        value.

        Args:
            value (Numerical): upper limit for the value

        Returns:
            Number: modified instance
        """
        self._f_lt = False
        self._f_lte = True
        self._m_ul = value
        return self

    def negative(self) -> Number:
        """
        Sets the upper limit for the value. The value has to be less than 0.

        Returns:
            Number: modified instance
        """
        return self.lt(0)

    def not_positive(self) -> Number:
        """
        Sets the upper limit for the value. The value has to be less than or equal to 0.

        Returns:
            Number: modified instance
        """
        return self.lte(0)

    def max(self, value: Numerical) -> Number:
        """
        Sets the upper limit for the value. The value has to be less than or equal to the specified
        value.

        Args:
            value (Numerical): upper limit for the value

        Returns:
            Number: modified instance
        """
        return self.lte(value)

    def _parse(self, obj: Numerical) -> Numerical:
        """
        Private function for parsing the value. This function is called by ``parse(..)`` and should
        not be called directly by the user.

        Throws:
            ValidationError: if the value could not be parsed or was invalid

        Args:
            obj (Numerical): value to parse

        Returns:
            Numerical: parsed destination value
        """
        # cast to integer, if required
        if self._f_integer:
            self._m_ll = int(self._m_ll) if self._m_ll is not None else None
            self._m_ul = int(self._m_ul) if self._m_ul is not None else None

        if self._f_lt and obj >= self._m_ul:
            raise ValidationError(f"object has to be < '{self._m_ul}', but is '{obj!r}'")

        elif self._f_lte and obj > self._m_ul:
            raise ValidationError(f"object has to be =<'{self._m_ul}', but is '{obj!r}'")

        if self._f_gt and obj <= self._m_ll:
            raise ValidationError(f"object has to be > '{self._m_ll}', but is '{obj!r}'")

        elif self._f_gte and obj < self._m_ll:
            raise ValidationError(f"object has to be >='{self._m_ll}', but is '{obj!r}'")

        return obj

    def parse(self, obj: Any) -> Numerical:
        # python handles bool as int, so we have to check for bool first
        if isinstance(obj, bool):
            # fallthrough on boolean values
            pass

        elif isinstance(obj, float):
            if self._f_integer:
                raise ValidationError(f"object has to be an integer, but is '{obj!r}'")
            return float(self._parse(obj))

        elif isinstance(obj, int):
            return int(self._parse(obj))

        raise ValidationError(f"object has to be a number, but is '{obj!r}'")

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[Numerical | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # if value is None, check if the value is nullable
            if value is None:
                if self._f_nullable:
                    return Some(None)
                raise ValidationError(f"key '{key}' cannot be None")
            # if key is found, just package ``parse(..)`` it into a Some
            return Some(self.parse(value))

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Array(ParasiteType[list[Any]]):
    """
    Parasite type for representing list values.

    Note:
        Please use ``p.array(...)`` instead of instantiating this class directly. ``p`` can be
        imported with::

            from parasite import p
            schema = p.array(...)
            ...
    """
    _m_element: ParasiteType | None = None   # The element type of the list.

    _f_optional: bool = False   # Whether the value is optional.
    _f_nullable: bool = False   # Whether the value can be None.

    _m_ul: int | None = None   # The upper limit of the list.
    _m_ll: int | None = None   # The lower limit of the list.

    def __init__(self, element: ParasiteType | None = None):
        """
        Args:
            element (ParasiteType | None): The element type of the list. Default: None

        Example usage::

            from parasite import p

            schema = p.array()
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse(["John", 1])  # -> ["John", 1]

            schema = p.array(p.string())
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse(["John", 1])  # -> ValidationError: element at index 1 is invalid: object has to be a string, but is 1
        """
        self._m_element = element

    def optional(self) -> Array:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.array() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required

            schema = p.obj({ "name": p.array().optional() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ })  # -> { }
        """
        self._f_optional = True
        return self

    def required(self) -> Array:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.array() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required

            schema = p.obj({ "name": p.array().required() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required
        """
        self._f_optional = False
        return self

    def nullable(self) -> Array:
        """
        Set the value to be nullable.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.array() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ "name": None })  # -> ValidationError: key 'name' cannot be None

            schema = p.obj({ "name": p.array().nullable() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ "name": None })  # -> { "name": None }
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Array:
        """
        Set the value to be non-nullable.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.array() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ "name": None })  # -> ValidationError: key 'name' cannot be None

            schema = p.obj({ "name": p.array().non_nullable() })
            schema.parse({ "name": ["John", "Doe"] })  # -> { "name": ["John", "Doe"] }
            schema.parse({ "name": None })  # -> ValidationError: key 'name' cannot be None
        """
        self._f_nullable = False
        return self

    def min(self, value: int) -> Array:
        """
        Set the minimum length of the list.

        Args:
            value (int): The minimum length of the list.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.array().min(2)
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse(["John"])  # -> ValidationError: list has to have at least 2 elements, but has 1
        """
        self._m_ll = value
        return self

    def max(self, value: int) -> Array:
        """
        Set the maximum length of the list.

        Args:
            value (int): The maximum length of the list.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.array().max(2)
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse(["John", "Doe", "Smith"])  # -> ValidationError: list has to have at most 2 elements, but has 3
        """
        self._m_ul = value
        return self

    def not_empty(self) -> Array:
        """
        Set the list to not be empty.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.array().not_empty()
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse([])  # -> ValidationError: list has to have at least 1 elements, but has 0
        """
        return self.min(1)

    def empty(self) -> Array:
        """
        Set the list to be empty.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.array().empty()
            schema.parse([])  # -> []
            schema.parse(["John", "Doe"])  # -> ValidationError: list has to have at most 0 elements, but has 2
        """
        return self.max(0)

    def length(self, value: int) -> Array:
        """
        Set the length of the list.

        Args:
            value (int): The length of the list.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.array().length(2)
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse(["John", "Doe", "Smith"])  # -> ValidationError: list has to have exactly 2 elements, but has 3
        """
        return self.min(value).max(value)

    def element(self, element: ParasiteType) -> Array:
        """
        Set the element type of the list.

        Args:
            element (ParasiteType[T]): The element type of the list.

        Returns:
            Array: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.array().element(p.string())
            schema.parse(["John", "Doe"])  # -> ["John", "Doe"]
            schema.parse(["John", 1])  # -> ValidationError: element at index 1 is invalid: object has to be a string, but is 1
        """
        self._m_element = element
        return self

    def parse(self, obj: Any) -> list[Any]:
        if not isinstance(obj, list):
            raise ValidationError(f"object has to be a list, but is '{obj!r}'")

        if self._m_ll is not None and len(obj) < self._m_ll:
            raise ValidationError(
                f"list has to have at least {self._m_ll} elements, but has {len(obj)}"
            )

        if self._m_ul is not None and len(obj) > self._m_ul:
            raise ValidationError(
                f"list has to have at most {self._m_ul} elements, but has {len(obj)}"
            )

        if self._m_element is not None:
            # create a cache to not overwrite the original list on error
            cache = []

            # parse each element individually
            for i, element in enumerate(obj):
                try:
                    # may throw a ValidationError exception
                    cache.append(self._m_element.parse(element))

                # handle ValidationError exceptions, if parsing fails
                except ValidationError as exc:
                    raise ValidationError(f"element at index {i} is invalid: {exc}") from exc

            # if the parsing was successful, overwrite the original list
            obj = cache

        return obj

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[list[Any] | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # if key is found, just package ``parse(..)`` it into a Some
            if value is not None:
                return Some(self.parse(value))

            # if value is None, check if the value is nullable
            if self._f_nullable:
                return Some(None)

            raise ValidationError(f"key '{key}' cannot be None")

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
import math
import re
from typing import Any, Optional, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Boolean(ParasiteType[bool]):
    """
    Parasite type for representing boolean values.

    Note:
        Please use ``p.boolean()`` instead of instantiating this class directly. ``p`` can be
        imported with::

            from parasite import p
            schema = p.boolean()
            ...
    """
    _f_optional: bool = False   # Whether the value is optional.
    _f_nullable: bool = False   # Whether the value can be None.
    _f_leaniant: bool = False   # Whether the value is leaniant.

    _m_leaniant: tuple[str, str] = (
        r"^(true|1|yes|y)$", r"^(false|0|no|n)$"
    )   # The regular expressions for the true and false value.
    _m_literal: bool | None = None   # The literal value of the boolean.

    def __init__(self) -> None:
        pass

    def optional(self) -> Boolean:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Returns:
            Boolean: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.boolean() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required

            schema = p.obj({ "name": p.boolean().optional() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ })  # -> { }
        """
        self._f_optional = True
        return self

    def required(self) -> Boolean:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Returns:
            Boolean: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.boolean() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required

            schema = p.obj({ "name": p.boolean().required() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required
        """
        self._f_optional = False
        return self

    def nullable(self) -> Boolean:
        """
        Set the value to be nullable.

        Returns:
            Boolean: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.boolean() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ "name": None })  # -> ValidationError: key 'name' cannot be None

            schema = p.obj({ "name": p.boolean().nullable() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ "name": None })  # -> { "name": None }
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Boolean:
        """
        Set the value to be not nullable.

        Returns:
            Boolean: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.boolean() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ "name": None })  # -> ValidationError: key 'name' cannot be None

            schema = p.obj({ "name": p.boolean().non_nullable() })
            schema.parse({ "name": True })  # -> { "name": True }
            schema.parse({ "name": None })  # -> ValidationError: key 'name' cannot be None
        """
        self._f_nullable = False
        return self

    def literal(self, value: bool) -> Boolean:
        """
        Set the literal value of the boolean.

        Args:
            value (bool): The literal value of the boolean.

        Returns:
            Boolean: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.boolean()
            schema.parse(True)  # -> True
            schema.parse(False)  # -> False

            schema = p.boolean().literal(True)
            schema.parse(True)  # -> True
            schema.parse(False)  # -> ValidationError: object has to be True, but is False
        """
        self._m_literal = value
        return self

    def leaniant(self, re_true: Optional[str] = None, re_false: Optional[str] = None) -> Boolean:
        """
        Set the value to be leaniant. This allows the value to be read from a string or a number.

        Args:
            re_true (Optional[str]): The regular expression for the true value. Default: None
            re_false (Optional[str]): The regular expression for the false value. Default: None

        Returns:
            Boolean: The updated instance of the class.

        Example usage::

            from parasite import p

            schema = p.boolean()
            schema.parse(True)  # -> True
            schema.parse("true")  # -> ValidationError: object has to be a boolean, but is 'true'
            schema.parse(1)  # -> ValidationError: object has to be a boolean, but is '1'

            schema = p.boolean().leaniant()
            schema.parse("true")  # -> True
            schema.parse("false")  # -> False
            schema.parse(1)  # -> True
        """
        self._f_leaniant = True

        if re_true is not None:
            self._m_leaniant = (re_true, self._m_leaniant[1])

        if re_false is not None:
            self._m_leaniant = (self._m_leaniant[0], re_false)

        return self

    def parse(self, obj: Any) -> bool:
        # if obj is already a boolean, return it
        if isinstance(obj, bool):
            pass

        # if obj is a string and leaniant mode is active, try to convert it to a boolean
        elif isinstance(obj, str) and self._f_leaniant:
            # if obj is a string, try to convert it to a boolean
            if re.match(self._m_leaniant[0], obj.lower()):
                obj = True

            elif re.match(self._m_leaniant[1], obj.lower()):
                obj = False

            else:
                # raise an error if the value could not be matched to any regex
                raise ValidationError(
                    f"object has to be regex (true: {self._m_leaniant[0]!r}, false: "
                    f"{self._m_leaniant[1]!r}) accepted boolean value, but is '{obj!r}'"
                )

        # if obj is a number, try to convert it to a boolean
        elif isinstance(obj, (int, float)) and self._f_leaniant:
            # first check if the number is an actual number
            if math.isnan(obj):
                obj = False

            elif int(obj) == 1:
                obj = True

            elif int(obj) == 0:
                obj = False

            else:
                raise ValidationError(f"object has to be 1 or 0, but is '{obj!r}'")

        else:
            # raise an error if the value could not be parsed
            raise ValidationError(f"object has to be a boolean, but is '{obj!r}'")

        # if literal is set, check if obj is the literal value
        if self._m_literal is not None and obj != self._m_literal:
            raise ValidationError(f"object has to be {self._m_literal}, but is {obj}")

        return obj

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[bool | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # if key is found, just package ``parse(..)`` it into a Some
            if value is not None:
                return Some(self.parse(value))

            # if value is None, check if the value is nullable
            if self._f_nullable:
                return Some(None)

            raise ValidationError(f"key '{key}' cannot be None")

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

# -- Library Imports --
from rusttypes.option import Option
from rusttypes.result import Err, Ok, Result

# -- Package Imports --
from parasite.errors import ValidationError

T = TypeVar("T")
"""Template type for the destination value."""

K = TypeVar("K")
"""Template type for the key in a dictionary."""

_NotFound = object()
"""Internal parasite value for not found keys."""


@dataclass
class ParasiteType(ABC, Generic[T]):
    """
    Abstract base class for parasite types. Parasite types are used to parse and validate values
    from a dictionary or standalone values (some options are only available for dictionaries).

    Inheritance:
        ABC, Generic[T]
    """

    @abstractmethod
    def parse(self, obj: Any) -> T:
        """
        Default method for parsing a value. This method should be overridden by subclasses.

        Throws:
            ValidationError: if the value could not be parsed or was invalid

        Args:
            obj (Any): value to parse

        Returns:
            T: parsed destination value
        """
        raise NotImplementedError

    @abstractmethod
    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[T | None]:
        """
        Default method for finding and parsing a value from a dictionary. This method should be
        overridden by subclasses. If the key is not found, the method should return ``Nil``.

        Throws:
            ValidationError: if the value could not be parsed or was invalid

        Args:
            parent (dict[K, Any]): dictionary to search for the key
            key (K): key to search for in the dictionary

        Returns:
            Option[T]: parsed destination value or ``Nil``
        """
        raise NotImplementedError

    def parse_safe(self, obj: Any) -> Result[T, ValidationError]:
        """
        Converts the result of :func:`parse` into a ``Result`` type. Should be used when safe
        parsing is required.

        Note:
            Will only catch :class:`parasite.errors.ValidationError` exceptions!!!

        Args:
            obj (Any): value to parse

        Returns:
            Result[T, ValidationError]: parsed destination value or an error
        """
        try:
            # may throw a ValidationError exception
            return Ok(self.parse(obj))

        # handle ValidationError exceptions, if parsing fails
        except ValidationError as exc:
            return Err(exc)

    def _find_and_parse_safe(self, parent: dict[K, Any],
                             key: K) -> Result[Option[T | None], ValidationError]:
        """
        Converts the result of :func:`_find_and_parse` into a ``Result`` type. Should be used when
        safe parsing is required.

        Note:
            Will only catch :class:`parasite.errors.ValidationError` exceptions!!!

        Args:
            parent (dict[K, Any]): dictionary to search for the key
            key (K): key to search for in the dictionary

        Returns:
            Result[T, ValidationError]: parsed destination value or an error
        """
        try:
            # may throw a ValidationError exception
            return Ok(self._find_and_parse(parent, key))

        # handle ValidationError exceptions, if parsing fails
        except ValidationError as exc:
            return Err(exc)



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Null(ParasiteType[None]):
    """
    Parasite type for representing None values.

    Note:
        Please use ``p.null()`` instead of instantiating this class directly. ``p`` can be
        imported with::

            from parasite import p
            schema = p.null()
            ...
    """
    _f_optional: bool = False   # Wether the value is optional

    def __init__(self) -> None:
        pass

    def optional(self) -> Null:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Returns:
            Null: modified instance

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.null() })
            schema.parse({ "name": None })  # -> { "name": None }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required

            schema = p.obj({ "name": p.null().optional() })
            schema.parse({ "name": None })  # -> { "name": None }
            schema.parse({ })  # -> { }
        """
        self._f_optional = True
        return self

    def required(self) -> Null:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Returns:
            Null: modified instance

        Example usage::

            from parasite import p

            schema = p.obj({ "name": p.null() })
            schema.parse({ "name": None })  # -> { "name": None }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required

            schema = p.obj({ "name": p.null().required() })
            schema.parse({ "name": None })  # -> { "name": None }
            schema.parse({ })  # -> ValidationError: key 'name' not found, but is required
        """
        self._f_optional = False
        return self

    def parse(self, obj: Any) -> None:
        # if obj is None, return None
        if obj is None:
            return None

        # else raise an error
        raise ValidationError(f"object has to be None, but is '{obj!r}'")

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[None]:
        # if key is found, just package ``parse(..)`` it into a Some
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            return Some(self.parse(value))

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
from re import Pattern
import re
from typing import Any, TypeVar
from enum import Enum

# -- Library Imports --
from rusttypes.option import Nil, Option, Some
import tracing

# -- Package Imports --
from parasite import _const
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class String(ParasiteType[str]):
    """
    Parasite type for representing string values.

    Inheritance:
        ParasiteType[str]

    Args:
        _f_optional (bool): Whether the value is optional. Default: False
        _f_nullable (bool): Whether the value can be None. Default: False
        _f_transform_before_parse (bool): Whether to transform the value before parsing.
        Default: False
        _f_trim (bool): Whether to trim the value. Default: False
        _f_to_lower (bool): Whether to convert the value to lowercase. Default: False
        _f_to_upper (bool): Whether to convert the value to uppercase. Default: False
        _m_regex_t (_RegexType): Type of regex to use for validation. Default: NONE
        _m_ul (int | None): Upper limit for the value. Default: None
        _m_ll (int | None): Lower limit for the value. Default: None
        _m_starts (str | None): String that the value must start with. Default: None
        _m_ends (str | None): String that the value must end with. Default: None
        _m_contains (str | None): String that the value must contain. Default: None
        _m_regex (Pattern | None): Compiled regex pattern to use for validation, if _m_regex_t is
        REGEX. Default: None
    """

    class _RegexType(Enum):
        """
        Enum for the different types of regexes that can be used to validate a string.
        """
        NONE = 0
        EMAIL = 1
        URL = 2
        UUID = 3
        CUID = 4
        CUID2 = 5
        ULID = 6
        IPV4 = 7
        IPV6 = 8
        REGEX = 9

    # flags
    _f_optional: bool = False
    _f_nullable: bool = False
    _f_transform_before_parse: bool = False

    # transformation flags
    _f_trim: bool = False
    _f_to_lower: bool = False
    _f_to_upper: bool = False

    _m_regex_t: _RegexType = _RegexType.NONE
    _m_ul: int | None = None
    _m_ll: int | None = None

    _m_starts: str | None = None
    _m_ends: str | None = None
    _m_contains: str | None = None
    _m_regex: Pattern | None = None

    def optional(self) -> String:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Returns:
            String: modified instance
        """
        self._f_optional = True
        return self

    def required(self) -> String:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Returns:
            String: modified instance
        """
        self._f_optional = False
        return self

    def nullable(self) -> String:
        """
        Makes the value nullable, when parsing with ``_find_and_parse(..)``. Has no effect on
        ``parse(..)``. Inverse of ``non_nullable(..)``.

        Returns:
            String: modified instance
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> String:
        """
        Makes the value non-nullable, when parsing with ``_find_and_parse(..)``. Has no effect on
        ``parse(..)``. Inverse of ``nullable(..)``. Default behavior.

        Returns:
            String: modified instance
        """
        self._f_nullable = False
        return self

    def trim(self) -> String:
        """
        Trims the value.

        Returns:
            String: modified instance
        """
        self._f_trim = True
        return self

    def to_lower(self) -> String:
        """
        Converts the value to lowercase.

        Returns:
            String: modified instance
        """
        self._f_to_lower = True
        return self

    def to_upper(self) -> String:
        """
        Converts the value to uppercase.

        Returns:
            String: modified instance
        """
        self._f_to_upper = True
        return self

    def transform_before_parse(self) -> String:
        """
        Transforms the value before parsing.

        Returns:
            String: modified instance
        """
        self._f_transform_before_parse = True
        return self

    def min(self, value: int) -> String:
        """
        Sets the lower length limit for the string value.

        Args:
            value (int): lower limit

        Returns:
            String: modified instance
        """
        self._m_ll = value
        return self

    def max(self, value: int) -> String:
        """
        Sets the upper length limit for the string value.

        Args:
            value (int): upper limit

        Returns:
            String: modified instance
        """
        self._m_ul = value
        return self

    def length(self, value: int) -> String:
        """
        Sets the lower and upper limit for the value.

        Args:
            value (int): length of the value

        Returns:
            String: modified instance
        """
        self._m_ll = value
        self._m_ul = value
        return self

    def not_empty(self) -> String:
        """
        Sets the lower limit for the value. The value has to be non-empty.

        Returns:
            String: modified instance
        """
        return self.min(1)

    def starts_with(self, value: str) -> String:
        """
        Sets the value that the string must start with.

        Args:
            value (str): string that the value must start with

        Returns:
            String: modified instance
        """
        self._m_starts = value
        return self

    def ends_with(self, value: str) -> String:
        """
        Sets the value that the string must end with.

        Args:
            value (str): string that the value must end with

        Returns:
            String: modified instance
        """
        self._m_ends = value
        return self

    def contains(self, value: str) -> String:
        """
        Sets the value that the string must contain.

        Args:
            value (str): string that the value must contain

        Returns:
            String: modified instance
        """
        self._m_contains = value
        return self

    def email(self) -> String:
        """
        Sets the regex type to email.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.EMAIL
        return self

    def url(self) -> String:
        """
        Sets the regex type to URL.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.URL
        return self

    def uuid(self) -> String:
        """
        Sets the regex type to UUID.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.UUID
        return self

    def cuid(self) -> String:
        """
        Sets the regex type to CUID.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.CUID
        return self

    def cuid2(self) -> String:
        """
        Sets the regex type to CUID2.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.CUID2
        return self

    def ulid(self) -> String:
        """
        Sets the regex type to ULID.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.ULID
        return self

    def ipv4(self) -> String:
        """
        Sets the regex type to IPv4.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.IPV4
        return self

    def ipv6(self) -> String:
        """
        Sets the regex type to IPv6.

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.IPV6
        return self

    def regex(self, value: Pattern) -> String:
        """
        Sets the regex pattern to use for validation.

        Args:
            value (Pattern): compiled regex pattern

        Returns:
            String: modified instance
        """
        self._m_regex_t = self._RegexType.REGEX
        self._m_regex = value
        return self

    def match(self, value: str) -> String:
        """
        Sets the regex pattern to use for validation.

        Args:
            value (str): regex pattern

        Returns:
            String: modified instance
        """
        return self.regex(re.compile(value))

    def _apply_transformations(self, value: str) -> str:
        """
        Applies the transformations on the value.

        Args:
            value (str): value to transform

        Returns:
            str: transformed value
        """
        if self._f_trim:
            value = value.strip()

        if self._f_to_lower:
            value = value.lower()

        if self._f_to_upper:
            value = value.upper()

        return value

    def _parse_bounds(self, value: str) -> str:
        """
        Parses the value with the length _constraints.

        Args:
            value (str): value to parse

        Returns:
            str: parsed value
        """
        if self._m_ll is not None and len(value) < self._m_ll:
            raise ValidationError(f"length of value '{value}' is less than {self._m_ll}")

        if self._m_ul is not None and len(value) > self._m_ul:
            raise ValidationError(f"length of value '{value}' is greater than {self._m_ul}")

        return value

    def _parse_basic(self, value: str) -> str:
        """
        Parses the value with the basic _constraints.

        Args:
            value (str): value to parse

        Returns:
            str: parsed value
        """
        if self._m_starts is not None and not value.startswith(self._m_starts):
            raise ValidationError(f"value '{value}' does not start with '{self._m_starts}'")

        if self._m_ends is not None and not value.endswith(self._m_ends):
            raise ValidationError(f"value '{value}' does not end with '{self._m_ends}'")

        if self._m_contains is not None and self._m_contains not in value:
            raise ValidationError(f"value '{value}' does not contain '{self._m_contains}'")

        return value

    def _parse_regex(self, value: str) -> str:
        """
        Parses the value with the regex _constraints.

        Args:
            value (str): value to parse

        Returns:
            str: parsed value
        """
        match self._m_regex_t:
            case self._RegexType.NONE:
                return value

            case self._RegexType.EMAIL:
                r = re.match(_const.RE_EMAIL, value)
                print(f"r: {r!r}, regex = {_const.RE_EMAIL!r}")

                if not _const.RE_EMAIL.match(value):
                    raise ValidationError(f"value '{value}' is not a valid email")
                return value

            case self._RegexType.URL:
                if not _const.RE_URL.match(value):
                    raise ValidationError(f"value '{value}' is not a valid URL")
                return value

            case self._RegexType.UUID:
                if not _const.RE_UUID.match(value):
                    raise ValidationError(f"value '{value}' is not a valid UUID")
                return value

            case self._RegexType.CUID:
                if not _const.RE_CUID.match(value):
                    raise ValidationError(f"value '{value}' is not a valid CUID")
                return value

            case self._RegexType.CUID2:
                if not _const.RE_CUID2.match(value):
                    raise ValidationError(f"value '{value}' is not a valid CUID2")
                return value

            case self._RegexType.ULID:
                if not _const.RE_ULID.match(value):
                    raise ValidationError(f"value '{value}' is not a valid ULID")
                return value

            case self._RegexType.IPV4:
                if not _const.RE_IPV4.match(value):
                    raise ValidationError(f"value '{value}' is not a valid IPv4")
                return value

            case self._RegexType.IPV6:
                if not _const.RE_IPV6.match(value):
                    raise ValidationError(f"value '{value}' is not a valid IPv6")
                return value

            case self._RegexType.REGEX:
                if not self._m_regex:
                    raise ValidationError("no regex pattern provided")
                if not self._m_regex.match(value):
                    raise ValidationError(f"value '{value}' does not match the regex pattern")
                return value

            case _:
                pass

        tracing.warn(f"unsupported regex type {self._m_regex_t!r}")
        return value

    def parse(self, obj: Any) -> str:
        if not isinstance(obj, str):
            raise ValidationError(f"expected a string, but got '{obj!r}'")

        if self._f_transform_before_parse:
            obj = self._apply_transformations(obj)

        obj = self._parse_bounds(obj)
        obj = self._parse_basic(obj)
        obj = self._parse_regex(obj)

        if not self._f_transform_before_parse:
            obj = self._apply_transformations(obj)

        return obj

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[str | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            if value is not None:
                # if key is found, just package ``parse(..)`` it into a Some
                return Some(self.parse(value))

            # if value is None, check if the value is nullable
            if self._f_nullable:
                return Some(None)

            raise ValidationError(f"key '{key}' cannot be None")

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
import copy
from dataclasses import dataclass, field
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound
from parasite.variant import Variant

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Object(ParasiteType[dict[Any, Any]]):
    """
    Parasite type for representing dictionary values.

    Note:
        Please use ``p.obj(...)`` instead of instantiating this class directly. ``p`` can be
        imported with::

            from parasite import p
            schema = p.obj({ ... })
            ...
    """
    # The items of the dictionary.s
    _m_items: dict[Any, ParasiteType] = field(default_factory = lambda: {})

    _f_optional: bool = False   # Whether the value is optional.
    _f_nullable: bool = False   # Whether the value can be None.
    # Whether the dictionary should be parsed and not allow any other keys to exist.
    _f_strict: bool = False
    # Whether the dictionary should be stripped of all keys that are not in the dictionary.
    _f_strip: bool = False

    def __init__(self, items: dict[Any, ParasiteType] = {}) -> None:
        """
        Args:
            items (dict[K, ParasiteType]): The schema of subitems of the dictionary. Default: {}.

        Example usage:
            You can create a dictionary schema by passing a dictionary of keys and their respective
            schemas. The following example shows how to create a schema for a dictionary with a
            the keys "name","age"::

                from parasite import p

                schema = p.obj({
                    "name": p.string(),
                    "age": p.number().integer().optional(),
                })
                schema2 = p.obj()

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "name": "John" })
                { "name": "John" }
                >>> schema.parse({ })
                ValidationError: key 'name' not found, but is required

                >>> schema2.parse({ })
                { }
                >>> schema2.parse({ "name": "John" })
                { "name": "John" }
        """
        self._m_items = items

    def optional(self) -> Object:
        """
        Makes the value optional, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`required`.

        Warning:
            This function has no effect if the value is parsed as a standalone value.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schemas::

                from parasite import p

                schema = p.obj({ "sub": p.obj().optional() })
                schema2 = p.obj({ "sub": p.obj() })

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "sub": { } })
                { "sub": { } }
                >>> schema.parse({ })
                { }

                >>> schema2.parse({ "sub": { } })
                { "sub": { } }
                >>> schema2.parse({ })
                ValidationError: key 'sub' not found, but is required
        """
        self._f_optional = True
        return self

    def required(self) -> Object:
        """
        Makes the value required, when parsing with :func:`_find_and_parse`. Has no effect on
        :func:`parse`. Inverse of :func:`optional`. Default behavior.

        Note:
            This function is default behavior for the class and therefore only has an effect if the
            function :func:`optional` may have been called before.

        Warning:
            This function has no effect if the value is parsed as a standalone value.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schemas::

                from parasite import p

                schema = p.obj({ "sub": p.obj().optional().required() })
                schema2 = p.obj({ "sub": p.obj() })

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "sub": { } })
                { "sub": { } }
                >>> schema.parse({ })
                ValidationError: key 'sub' not found, but is required

                >>> schema2.parse({ "sub": { } })
                { "sub": { } }
                >>> schema2.parse({ })
                ValidationError: key 'sub' not found, but is required
        """
        self._f_optional = False
        return self

    def nullable(self) -> Object:
        """
        Set the value to be nullable.

        Warning:
            This function has no effect if the value is parsed as a standalone value.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schemas::

                from parasite import p

                schema = p.obj({ "sub": p.obj().nullable() })
                schema2 = p.obj({ "sub": p.obj() })

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "sub": { } })
                { "sub": { } }
                >>> schema.parse({ "sub": None })
                { "sub": None }

                >>> schema2.parse({ "sub": { } })
                { "sub": { } }
                >>> schema2.parse({ "sub": None })
                ValidationError: object has to be a dictionary, but is 'None'


        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Object:
        """
        Set the value to be non-nullable. This function only has an effect if the value is parsed
        inside a dictionary.

        Note:
            This function is default behavior for the class and therefore only has an effect if the
            function :func:`nullable` may have been called before.

        Warning:
            This function has no effect if the value is parsed as a standalone value.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schemas::

                from parasite import p

                schema = p.obj({ "sub": p.obj().nullable().non_nullable() })
                schema2 = p.obj({ "sub": p.obj() })

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "sub": { } })
                { "sub": { } }
                >>> schema.parse({ "sub": None })
                ValidationError: object has to be a dictionary, but is 'None'

                >>> schema2.parse({ "sub": { } })
                { "sub": { } }
                >>> schema2.parse({ "sub": None })
                ValidationError: object has to be a dictionary, but is 'None'
        """
        self._f_nullable = False
        return self

    def strict(self) -> Object:
        """
        Set the dictionary to be strict. This function ensures that only the keys in the schema are
        allowed in the dictionary. If this function is used :func:`strip` has no effect.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schemas::

                from parasite import p

                schema = p.obj({ "name": p.string() }).strict()
                schema2 = p.obj({ "name": p.string() })

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "name": "John" })
                { "name": "John" }
                >>> schema.parse({ "name": "John", "age": 20 })
                ValidationError: object has the key 'age', but is not allowed to

                >>> schema2.parse({ "name": "John" })
                { "name": "John" }
                >>> schema2.parse({ "name": "John", "age": 20 })
                { "name": "John", "age": 20 }
        """
        self._f_strict = True
        return self

    def strip(self) -> Object:
        """
        Set the dictionary to be stripped. This will remove all keys that are not in the schema. In
        constrast to :func:`strict` this function does not raise an error if a key is not in the
        schema.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schemas::

                from parasite import p

                schema = p.obj({ "name": p.string() }).strip()
                schema2 = p.obj({ "name": p.string() })

            The resulting schemas will parse the following objects::

                >>> schema.parse({ "name": "John" })
                { "name": "John" }
                >>> schema.parse({ "name": "John", "age": 20 })
                { "name": "John" }

                >>> schema2.parse({ "name": "John" })
                { "name": "John" }
                >>> schema2.parse({ "name": "John", "age": 20 })
                { "name": "John", "age": 20 }
        """
        self._f_strip = True
        return self

    def extend(self, other: Object) -> Object:
        """
        Extend the dictionary with another dictionary. This will overwrite all values of the current
        dictionary with the values of the other dictionary if the key exists in both dictionaries.
        If you want to merge the dictionaries, use :func:`merge`.

        Args:
            other (Object): The dictionary to extend with.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following two schemas::

                from parasite import p

                schema = p.obj({ "name": p.string(), "age": p.string() })
                schema2 = p.obj({ "age": p.number().integer() })

            If we extend the two schemas, the resulting schema will be::

                # -> schema.extend(schema2)

                p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                })

            The resulting schema will parse the following objects::

                >>> schema.parse({ "name": "John", "age": 20 })
                { "name": "John", "age": 20 }

                >>> schema.parse({ "name": "John", "age": "20" })
                ValidationError: key 'age' has to be an integer, but is '20'
        """
        if not isinstance(other, Object):
            raise ValidationError(f"object has to be a dictionary, but is '{other!r}'")

        self._m_items.update(other._m_items)
        return self

    def merge(self, other: Object) -> Object:
        """
        Merge the dictionary with another dictionary. This tries to merge the values of the current
        dictionary with the values of the other dictionary. If the key exists in both dictionaries
        the value of the current dictionary will not be overwritten, but the values will be merged.
        This results in the following behavior:

        - If both values are objects, they will be merged into a single dictionary with :func:`merge`.
        - If both values are variants, they will be merged into a single variant.
        - If one value is a variant and the other is not, the value will be added to the variant.
        - If both values are something else, they will be merged into a new variant.

        Args:
            other (Object): The dictionary to merge with.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following two schemas::

                from parasite import p

                schema = p.obj({ "name": p.string(), "age": p.string() })
                schema2 = p.obj({ "age": p.number().integer() })

            If we merge the two schemas, the resulting schema will be::

                # -> schema.merge(schema2)

                p.obj({
                    "name": p.string(),
                    "age": p.variant([
                         p.string(),
                         p.number().integer(),
                     ])
                })

            The resulting schema will parse the following objects::

                >>> schema.parse({ "name": "John", "age": 20 })
                { "name": "John", "age": 20 }

                >>> schema.parse({ "name": "John", "age": "20" })
                { "name": "John", "age": "20" }
        """
        for key, value in other._m_items.items():

            # If the key is in the dictionary, merge the values.
            if key in self._m_items:
                # if both src and dest are objects, merge them
                if isinstance(value, Object) and isinstance(self._m_items[key], Object):
                    self._m_items[key].merge(value)

                # if both src and dest are arrays, merge them
                elif isinstance(value, Variant) and isinstance(self._m_items[key], Variant):
                    for variant in value._m_variants:
                        self._m_items[key].add_variant(variant)

                # if dest is already a variant, add the value to it
                elif isinstance(self._m_items[key], Variant):
                    self._m_items[key].add_variant(value)

                # else, create a new variant and add both values
                else:
                    org = self._m_items[key]
                    self._m_items[key] = Variant().add_variant(org).add_variant(value)

            # If the key is not in the dictionary, add it.
            else:
                self._m_items[key] = value

        return self

    def pick(self, keys: list) -> Object:
        """
        Pick only the keys from the object. This will create a new :class:`Object` with only the
        keys found in the list. If a key is not found in the object, it will raise a KeyError. If
        you want to ignore keys that are not found, use :func:`pick_safe`.

        Args:
            keys (list[K]): The keys to pick.

        Returns:
            Object: New instance of the class with only the picked keys.

        Raises:
            KeyError: If a key is not found in the object.

        Example usage:
            Lets assume we have the following schema::

                from parasite import p

                schema = p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                    "city": p.string(),
                }).strict()

            If we pick the keys "name" and "age", the resulting schema will be::

                # -> schema = schema.pick(["name", "age"])

                p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                }).strict()

            The resulting schema will parse the following objects::

                >>> schema.parse({ "name": "John", "age": 20 })
                { "name": "John", "age": 20 }

                >>> schema.parse({ "name": "John", "age": 20, "city": "New York" })
                ValidationError: object has the key 'city', but is not allowed to
        """
        new_obj = copy.deepcopy(self)
        new_obj._m_items = {key: self._m_items[key] for key in keys}
        return new_obj

    def pick_safe(self, keys: list) -> Object:
        """
        Pick only the keys from the object. This will create a new :class:`Object` with only the
        keys found in the list. If a key is not found in the object, it will be ignored. If you want
        to raise an error if a key is not found, use :func:`pick`.

        Args:
            keys (list[K]): The keys to pick.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schema::

                from parasite import p

                schema = p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                    "city": p.string(),
                }).strict()

            If we pick the keys "name" and "age", the resulting schema will be::

                # -> schema = schema.pick_safe(["name", "age"])

                p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                }).strict()

            The resulting schema will parse the following objects::

                >>> schema.parse({ "name": "John", "age": 20 })
                { "name": "John", "age": 20 }

                >>> schema.parse({ "name": "John", "age": 20, "city": "New York" })
                { "name": "John", "age": 20 }
        """
        new_obj = copy.deepcopy(self)
        new_obj._m_items = {key: self._m_items[key] for key in keys if key in self._m_items}
        return new_obj

    def omit(self, keys: list) -> Object:
        """
        Omit the keys from the object. This will create a new :class:`Object` with all keys except
        the ones found in the list.

        Args:
            keys (list[K]): The keys to omit.

        Returns:
            Object: New instance of the class with the omitted keys.

        Example usage:
            Lets assume we have the following schema::

                from parasite import p

                schema = p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                    "city": p.string(),
                }).strict()

            If we omit the keys "name" and "age", the resulting schema will be::

                # -> schema = schema.omit(["name", "age"])

                p.obj({
                    "city": p.string(),
                }).strict()

            The resulting schema will parse the following objects::

                >>> schema.parse({ "name": "John", "age": 20, "city": "New York" })
                ValidationError: object has the key 'name', but is not allowed to

                >>> schema.parse({ "city": "New York" })
                { "city": "New York" }
        """
        new_obj = copy.deepcopy(self)
        new_obj._m_items = {key: value for key, value in self._m_items.items() if key not in keys}
        return new_obj

    def add_item(self, key: Any, item: ParasiteType) -> Object:
        """
        Add an item to the object. This will also overwrite the item if it already exists.

        Args:
            key (K): The key of the item.
            item (ParasiteType): The item to add.

        Returns:
            Object: The updated instance of the class.

        Example usage:
            Lets assume we have the following schema::

                from parasite import p

                schema = p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                }).strict()

            If we add the key "city" with a string schema, the resulting schema will be::

                # -> schema.add_item("city", p.string())

                p.obj({
                    "name": p.string(),
                    "age": p.number().integer(),
                    "city": p.string(),
                }).strict()

            The resulting schema will parse the following objects::

                >>> schema.parse({ "name": "John", "age": 20, "city": "New York" })
                { "name": "John", "age": 20, "city": "New York" }

                >>> schema.parse({ "name": "John", "age": 20 })
                ValidationError: key 'city' not found, but is required
        """
        self._m_items[key] = item
        return self

    def parse(self, obj: Any) -> dict[Any, Any]:
        if not isinstance(obj, dict):
            raise ValidationError(f"object has to be a dictionary, but is '{obj!r}'")

        # If the dictionary should be strict, check if all keys are allowed.
        if self._f_strict:
            for key in obj.keys():
                if key not in self._m_items:
                    raise ValidationError(f"object has the key '{key}', but is not allowed to")

        # If the dictionary should be stripped, strip it.
        if self._f_strip:
            obj = {key: value for key, value in obj.items() if key in self._m_items.keys()}

        # Parse the dictionary.
        for key, item in self._m_items.items():
            item._find_and_parse(obj, key).map(lambda x: obj.update({key: x}))

        return obj

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[dict[K, Any] | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # If key is found, just package ``parse(..)`` it into a Some.
            if value is not None:
                return Some(self.parse(value))

            # If value is None, check if the value is nullable.
            if self._f_nullable:
                return Some(None)

        # If the value is optional, return a Nil.
        if self._f_optional:
            return Nil

        # If the value is required, raise an error.
        raise ValidationError(f"object has to be a dictionary, but is '{value!r}'")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Never(ParasiteType[None]):
    """
    Parasite type for representing never values.

    Note:
        Please use ``p.never()`` instead of instantiating this class directly. ``p`` can be
        imported with::

            from parasite import p
            schema = p.never()
            ...
    """

    def parse(self, obj: Any) -> None:
        # always raise an error, as this type can never be parsed
        raise ValidationError("this type can never be parsed")

    def _find_and_parse(self, parent: dict[K, Any], key: K) -> Option[None]:
        # if key is found, raise an error
        if parent.get(key, _NotFound) is not _NotFound:
            raise ValidationError(f"key '{key}' found, but this type can never be parsed")

        return Nil



# -- STL Imports --
import re

RE_EMAIL = re.compile(
    r"^[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z"
    r"0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$",
)
"""
Regex pattern for email validation as per RFC2822 standards.

Attribution:
    - https://regexr.com/2rhq7, by Tripleaxis (from .NET helpfiles)
"""

RE_URL = re.compile(r"^((\w+:\/\/)[-a-zA-Z0-9:@;?&=\/%\+\.\*!'\(\),\$_\{\}\^~\[\]`#|]+)$",)
"""
Regex pattern for URL/URI validation.

Attribution:
    - https://regexr.com/2ri7q, by Gabriel Mariani
"""

RE_UUID = re.compile(
    r'^[0-9A-Za-z]{8}-[0-9A-Za-z]{4}-4[0-9A-Za-z]{3}-[89ABab][0-9A-Za-z]{3}-[0-9A-Za-z]{12}$'
)
"""Regex pattern validation for UUID v4 as per RFC9562.

Attribution:
    - http://en.wikipedia.org/wiki/Universally_unique_identifier#Version_4_.28random.29
    - https://regexr.com/39f77, by clayzermk1
"""

# Copyright (c) 2022 Colin McDonnell
# All rights reserved.
#
# Original Project: https://github.com/colinhacks/zod
# License File: licenses/zod_mit.txt
RE_CUID = re.compile(r'^c[^\\s-]{8,}$')
"""Regex pattern for CUID validation."""

# Copyright (c) 2022 Colin McDonnell
# All rights reserved.
#
# Original Project: https://github.com/colinhacks/zod
# License File: licenses/zod_mit.txt
RE_CUID2 = re.compile(r'^[a-z][a-z0-9]*$')
"""Regex pattern for CUID2 validation."""

# Copyright (c) 2022 Colin McDonnell
# All rights reserved.
#
# Original Project: https://github.com/colinhacks/zod
# License File: licenses/zod_mit.txt
RE_ULID = re.compile(r'^[0-9A-HJKMNP-TV-Z]{26}$')
"""Regex pattern for ULID validation."""

# Copyright (c) 2022 Colin McDonnell
# All rights reserved.
#
# Original Project: https://github.com/colinhacks/zod
# License File: licenses/zod_mit.txt
RE_IPV4 = re.compile(
    r"^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])$"
)
"""Regex pattern for IPv4 validation."""

# Copyright (c) 2022 Colin McDonnell
# All rights reserved.
#
# Original Project: https://github.com/colinhacks/zod
# License File: licenses/zod_mit.txt
RE_IPV6 = re.compile(
    r'^(([a-f0-9]{1,4}:){7}|::([a-f0-9]{1,4}:){0,6}|([a-f0-9]{1,4}:){1}:([a-f0-9]{1,4}:){0,5}|'
    r'([a-f0-9]{1,4}:){2}:([a-f0-9]{1,4}:){0,4}|([a-f0-9]{1,4}:){3}:([a-f0-9]{1,4}:){0,3}|'
    r'([a-f0-9]{1,4}:){4}:([a-f0-9]{1,4}:){0,2}|([a-f0-9]{1,4}:){5}:([a-f0-9]{1,4}:){0,1})([a-f0-9]'
    r'{1,4}|(((25[0-5])|(2[0-4][0-9])|(1[0-9]{2})|([0-9]{1,2}))\\.){3}((25[0-5])|(2[0-4][0-9])|'
    r'(1[0-9]{2})|([0-9]{1,2})))$'
)
"""Regex pattern for IPv6 validation."""
